/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#ifndef __GDAM_CHANNEL_H_
#define __GDAM_CHANNEL_H_

#ifdef __cplusplus 
extern "C" { 
#endif

typedef struct _GdamChannel GdamChannel;
typedef struct _GdamChannelEventHandler GdamChannelEventHandler;

#include <gdam/gdambuffer.h>
#include <gdam/gdamarg.h>
#include "gdamrequest.h"


/*
 *   ____     _                  ____ _                            _
 *  / ___| __| | __ _ _ __ ___  / ___| |__   __ _ _ __  _ __   ___| |
 * | |  _ / _` |/ _` | '_ ` _ \| |   | '_ \ / _` | '_ \| '_ \ / _ \ |
 * | |_| | (_| | (_| | | | | | | |___| | | | (_| | | | | | | |  __/ |
 *  \____|\__,_|\__,_|_| |_| |_|\____|_| |_|\__,_|_| |_|_| |_|\___|_|
 * 
 * 
 * The client-side of a connection to the gdam-server.
 */
struct _GdamChannel {
	GdamBuffer		outgoing;

	/* Incoming data first goes here, then gets processed into
	 * responses which are handled by `first_pending'. */
	GdamBuffer		incoming;

	/* Requests which haven't been copied into outgoing. */
	GSList*			first_request;
	GSList*			last_request;

	/* Requests which have been sent (at least to `outgoing').
	 * These requests cannot be reordered. */
	GSList*			first_pending;
	GSList*			last_pending;

	/* num_pending is the length of the first/last_pending list.
	 * max_pending is used to determine if we should write out
	 *             first_request into outgoing and move the request
	 *             to last_pending...
	 *
	 *             Please note that max_pending is not relevant
	 *             to the number of requests you may dispatch-
	 *             that is unlimited.  (This is just a tweak
	 *             so we jam more data down a slow connection.)
	 */
	int			num_pending;
	int			max_pending;

	/* TODO:  "events" from the server must be handled. */

	gboolean              (*auth_handler)    (GdamChannel* channel,
	                                          GdamBuffer*  auth_response);

	/* Our `pollfd' which allows us to find out what we can
	 * do with `channel'. */
	GPollFD			poll_fd;

	/* Our mainloop source id. */
	guint			source_id;

	/* Map from event-connection id to event information
	 * (internal use only) */
	GHashTable*		event_to_info_map;

	int			ref_count;

	/* Whether the connection to the server is in the middle of
	 * of connecting.
	 * Don't read or write from `connecting' sockets... */
	unsigned		is_connecting : 1,

	/* Whether we should quit the main loop if
	 * connection fails... */
				quit_main : 1;
};


struct _GdamChannelEventHandler {
	guint			connection_id;

	GHookFunc		callback;
	gpointer		user_data;


	/* Returns TRUE if it is done and won't be called again... */
	gboolean              (*marshal)(GdamChannelEventHandler*,
	                                 GdamBuffer*,
					 guint connection_id);

};


/* Usage message that can be displayed to users... */
extern const char* gdam_channel_usage_message;

gboolean     gdam_channel_init                    (int*        argc,
                                                   char***     argv,
						   GdamChannel** channel_out);
GdamChannel* gdam_channel_new                     ();
GdamChannel* gdam_channel_new_full                (const char* addr_spec,
                                                   GdamBuffer* auth_info);


/* Call unref on any source you "new". */
void         gdam_channel_unref                   (GdamChannel* channel);
void         gdam_channel_ref                     (GdamChannel* channel);

/* Attach to the main loop... you have to do this for
 * events to get processed. */
void         gdam_channel_attach                  (GdamChannel* channel);
void         gdam_channel_detach                  (GdamChannel* channel);

/* Delete us (safe whether or not attached). */
void         gdam_channel_destroy                 (GdamChannel* channel);

void         gdam_channel_register_handler        (GdamChannel* channel,
                                                   GdamChannelEventHandler*);
void         gdam_channel_unregister_handler      (GdamChannel* channel,
                                                   guint        connection_id);

GdamChannelEventHandler*
             gdam_channel_simple_handler_new      (guint connection_id,
						   gboolean one_shot,
	                                           GHookFunc handler_func,
						   gpointer handler_data);
GdamChannelEventHandler*
             gdam_channel_int_handler_new         (guint        connection_id,
						   gboolean     one_shot,
	                                           GdamRequestGotIntFunc,
						   gpointer     handler_data);
GdamChannelEventHandler*
             gdam_channel_time_handler_new        (guint        connection_id,
						   gboolean     one_shot,
	                                           GdamRequestGotTimeFunc,
						   gpointer     handler_data);
GdamChannelEventHandler*
             gdam_channel_2int_handler_new        (guint        connection_id,
						   gboolean     one_shot,
	                                           GdamRequestGot2IntFunc,
						   gpointer     handler_data);

/* These are trivial functions built around the asynchronous core... */

guint        gdam_channel_get_default_device      (GdamChannel* channel);
guint        gdam_channel_create_sink             (GdamChannel* channel,
                                                   const char*  device_name);
guint        gdam_channel_create_empty_mix        (GdamChannel* channel);

guint        gdam_channel_create_source_generic_v (GdamChannel* channel,
                                                   const char*  type,
						   int          num_args,
						   GdamArg*     args);
guint        gdam_channel_create_source_generic   (GdamChannel* channel,
                                                   const char*  type,
						   int          num_args,
						   ...);
guint        gdam_channel_source_set_arg          (GdamChannel* channel,
                                                   int          source_id,
						   /* next is arg_name,
						    * arg_type,
						    * arg_value */
						   ...);
guint        gdam_channel_create_mp3              (GdamChannel* channel,
                                                   const char*  filename,
						   GdamTime     start,
						   GdamTime     length);
gboolean     gdam_channel_mix_add                 (GdamChannel* channel,
                                                   guint        mix_id,
						   guint        subsource_id,
						   GdamTime     when);
						   /* if when==-1,
						         add immediately.*/


gboolean     gdam_channel_synchronize             (GdamChannel* channel,
                                                   int          mobile_id,
                                                   int          fixed_id,
					           gdouble      mobile_bpm,
					           gdouble      mobile_start,
					           gdouble      fixed_bpm,
					           gdouble      fixed_start);
gboolean     gdam_channel_sync_preview            (GdamChannel* channel,
                                                   int          mobile_id,
                                                   int          fixed_id,
					           gdouble      mobile_bpm,
					           gdouble      mobile_start,
					           gdouble      fixed_bpm,
					           gdouble      fixed_start,
						   GdamTime*    time_out);


guint        gdam_channel_start_caching           (GdamChannel* channel,
                                                   int          to_drain_id,
						   GHookFunc    done_caching,
						   gpointer     done_data);
gboolean     gdam_channel_uncache                 (GdamChannel* channel,
                                                   guint        buffer_id);

gboolean     gdam_channel_detach_source           (GdamChannel* channel,
                                                   guint        source_id);
gboolean     gdam_channel_feed_sink               (GdamChannel* channel,
                                                   guint        device_id,
                                                   guint        source_id);
gboolean     gdam_channel_seek_relative           (GdamChannel* channel,
                                                   guint        source_id,
						   GdamTime     when);
#if 0
gboolean     gdam_channel_set_sequencer           (GdamChannel* channel,
                                                   guint        sequence_id,
                                                   guint        sequence_len,
						   int          num_elements,
			  const GdamSoundSpecSequencerElement*  elements);
				                   
GdamSoundSpec* gdam_channel_get_sound_spec        (GdamChannel* channel,
                                                   guint        source_id,
					           gboolean     can_recurse);
#endif
gboolean     gdam_channel_get_server_side_int     (GdamChannel* channel,
                                                   const char*  category,
						   const char*  variable,
						   gint*        output);
guint        gdam_channel_subscribe_source_ended  (GdamChannel* channel,
                                                   guint        source_id,
						   GHookFunc    on_ended,
						   gpointer     on_ended_data);
guint        gdam_channel_subscribe_source_timer  (GdamChannel* channel,
                                                   guint        source_id,
						   int          period,
						   GdamRequestGotTimeFunc,
						   gpointer     periodic_data);
gboolean     gdam_channel_unsubscribe             (GdamChannel* channel,
                                                   guint        subscription);


gboolean     gdam_channel_append_request          (GdamChannel* channel,
                                                   GdamRequest* request);
gboolean     gdam_channel_prepend_request         (GdamChannel* channel,
                                                   GdamRequest* request);

/* Process a request (synchronously). */
gboolean     gdam_channel_process_request         (GdamChannel* channel,
                                                   GdamRequest* request);

/* If you allocated a request waitable but don't want to wait
 * after all. */
void         gdam_channel_disown_request          (GdamChannel* channel,
                                                   GdamRequest* request);

#ifdef __cplusplus 
}
#endif

#endif
