/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamarg.h"
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <gtk/gtksignal.h>

typedef struct _GdamArgInfo GdamArgInfo;

typedef gboolean   (*ArgumentParser) (GdamBuffer*   buffer,
                                      GdamArg*      arg);
typedef void       (*ArgumentValist) (va_list*      args,
                                      GdamArg*      arg);
typedef void       (*ArgumentWriter) (GdamBuffer*   buffer,
                                      const GdamArg* arg);
typedef void       (*ArgumentClearer)(GdamArg*      arg);

/* Parsing various fundamental types from buffers. */
static gboolean gdam_arg_parse_int(GdamBuffer* buffer, GdamArg* arg)
{
	return gdam_buffer_parse_int(buffer, &arg->d.int_data);
}
static gboolean gdam_arg_parse_int64(GdamBuffer* buffer, GdamArg* arg)
{
	return gdam_buffer_parse_int64(buffer, &arg->d.int64_data);
}
static gboolean gdam_arg_parse_string(GdamBuffer* buffer, GdamArg* arg)
{
	char* string = gdam_buffer_parse_string0(buffer);
	arg->d.string_data = string;
	return string ? TRUE : FALSE;
}
static gboolean gdam_arg_parse_double(GdamBuffer* buffer, GdamArg* arg)
{
	return gdam_buffer_parse_double(buffer, &arg->d.double_data);
}

static gboolean gdam_arg_parse_int_array(GdamBuffer* buffer, GdamArg* arg)
{
	int length, i;
	int *array;
	if (!gdam_buffer_parse_int(buffer, &length))
		return FALSE;
	array = g_new(int, length);
	for (i = 0; i < length; i++) {
		if (!gdam_buffer_parse_int(buffer, array + i)) {
			g_free(array);
			return FALSE;
		}
	}
	GDAM_ARG_INT_ARRAY(arg).length = length;
	GDAM_ARG_INT_ARRAY(arg).array = array;
	return TRUE;
}
static gboolean gdam_arg_parse_int64_array(GdamBuffer* buffer, GdamArg* arg)
{
	int length, i;
	gint64 *array;
	if (!gdam_buffer_parse_int(buffer, &length))
		return FALSE;
	array = g_new(gint64, length);
	for (i = 0; i < length; i++)
		if (!gdam_buffer_parse_int64(buffer, array + i)) {
			g_free(array);
			return FALSE;
		}
	GDAM_ARG_INT64_ARRAY(arg).length = length;
	GDAM_ARG_INT64_ARRAY(arg).array = array;
	return TRUE;
}
static gboolean gdam_arg_parse_string_array(GdamBuffer* buffer, GdamArg* arg)
{
	int length, i;
	char** array;
	if (!gdam_buffer_parse_int(buffer, &length))
		return FALSE;
	array = g_new(char*, length);
	for (i = 0; i < length; i++)
		if (!(array[i]=gdam_buffer_parse_string0(buffer))) {
			int j;
			for (j = 0; j < i; j++)
				g_free(array[j]);
			g_free(array);
			return FALSE;
		}
	GDAM_ARG_STRING_ARRAY(arg).length = length;
	GDAM_ARG_STRING_ARRAY(arg).array = array;
	return TRUE;
}
static gboolean gdam_arg_parse_double_array(GdamBuffer* buffer, GdamArg* arg)
{
	int length, i;
	gdouble *array;
	if (!gdam_buffer_parse_int(buffer, &length))
		return FALSE;
	array = g_new(gdouble, length);
	for (i = 0; i < length; i++)
		if (!gdam_buffer_parse_double(buffer, array + i)) {
			g_free(array);
			return FALSE;
		}
	GDAM_ARG_DOUBLE_ARRAY(arg).length = length;
	GDAM_ARG_DOUBLE_ARRAY(arg).array = array;
	return TRUE;
}

/* Parsing various fundamental types from varargs. */
static void gdam_arg_valist_int(va_list* args, GdamArg* arg)
{
	arg->d.int_data = va_arg(*args, int);
}
static void gdam_arg_valist_int64(va_list* args, GdamArg* arg)
{
	arg->d.int64_data = va_arg(*args, gint64);
}
static void gdam_arg_valist_string(va_list* args, GdamArg* arg)
{
	const char* string = va_arg(*args, const char*);
	arg->d.string_data = g_strdup(string);
}
static void gdam_arg_valist_double(va_list* args, GdamArg* arg)
{
	arg->d.double_data = va_arg(*args, gdouble);
}
static void gdam_arg_valist_int_array(va_list* args, GdamArg* arg)
{
	int length; int* orig;
	int* copy;
	length = va_arg(*args, int);
	orig = va_arg(*args, int*);
	copy = g_new(int, length);
	memcpy(copy, orig, sizeof(int) * length);
	GDAM_ARG_INT_ARRAY(arg).length = length;
	GDAM_ARG_INT_ARRAY(arg).array = copy;
}
static void gdam_arg_valist_int64_array(va_list* args, GdamArg* arg)
{
	int length; gint64* orig;
	gint64* copy;
	length = va_arg(*args, int);
	orig = va_arg(*args, gint64*);
	copy = g_new(gint64, length);
	memcpy(copy, orig, sizeof(gint64) * length);
	GDAM_ARG_INT64_ARRAY(arg).length = length;
	GDAM_ARG_INT64_ARRAY(arg).array = copy;
}
static void gdam_arg_valist_string_array(va_list* args, GdamArg* arg)
{
	int length; const char* const* orig;
	char** copy;
	int i;
	length = va_arg(*args, int);
	orig = va_arg(*args, const char* const*);
	copy = g_new(char*, length);
	for (i = 0; i < length; i++)
		copy[i] = g_strdup(orig[i]);
	GDAM_ARG_STRING_ARRAY(arg).length = length;
	GDAM_ARG_STRING_ARRAY(arg).array = copy;
}

static void gdam_arg_valist_double_array(va_list* args, GdamArg* arg)
{
	int length; gdouble* orig;
	gdouble* copy;
	length = va_arg(*args, int);
	orig = va_arg(*args, gdouble*);
	copy = g_new(gdouble, length);
	memcpy(copy, orig, sizeof(gdouble) * length);
	GDAM_ARG_DOUBLE_ARRAY(arg).length = length;
	GDAM_ARG_DOUBLE_ARRAY(arg).array = copy;
}

/* Writing various fundamental types into buffers. */
static void gdam_arg_writer_int(GdamBuffer* buffer, const GdamArg* arg)
{
	gdam_buffer_append_int(buffer, arg->d.int_data);
}
static void gdam_arg_writer_int64(GdamBuffer* buffer, const GdamArg* arg)
{
	gdam_buffer_append_int64(buffer, arg->d.int64_data);
}
static void gdam_arg_writer_string(GdamBuffer* buffer, const GdamArg* arg)
{
	gdam_buffer_append_string0(buffer, arg->d.string_data);
}
static void gdam_arg_writer_double(GdamBuffer* buffer, const GdamArg* arg)
{
	gdam_buffer_append_double(buffer, arg->d.double_data);
}
static void gdam_arg_writer_int_array(GdamBuffer* buffer, const GdamArg* arg)
{
	int length = GDAM_ARG_INT_ARRAY(arg).length, i;
	int* array = GDAM_ARG_INT_ARRAY(arg).array;
	gdam_buffer_append_int(buffer, length);
	for (i = 0; i < length; i++)
		gdam_buffer_append_int(buffer, array[i]);
}
static void gdam_arg_writer_int64_array(GdamBuffer* buffer, const GdamArg* arg)
{
	int length = GDAM_ARG_INT64_ARRAY(arg).length, i;
	gdam_buffer_append_int(buffer, length);
	for (i = 0; i < length; i++)
		gdam_buffer_append_int64(buffer,
		                         GDAM_ARG_INT64_ARRAY(arg).array[i]);
}
static void gdam_arg_writer_string_array(GdamBuffer* buffer, const GdamArg* arg)
{
	int length = GDAM_ARG_STRING_ARRAY(arg).length, i;
	gdam_buffer_append_int(buffer, length);
	for (i = 0; i < length; i++)
		gdam_buffer_append_string0(buffer,
		                           GDAM_ARG_STRING_ARRAY(arg).array[i]);
}
static void gdam_arg_writer_double_array(GdamBuffer* buffer, const GdamArg* arg)
{
	int length = GDAM_ARG_DOUBLE_ARRAY(arg).length, i;
	gdam_buffer_append_int(buffer, length);
	for (i = 0; i < length; i++)
		gdam_buffer_append_double(buffer,
		                          GDAM_ARG_DOUBLE_ARRAY(arg).array[i]);
}

/* Things to free various fundamental types. */
static void gdam_arg_clearer_string(GdamArg* arg)
{
	g_free(arg->d.string_data);
}
static void gdam_arg_clearer_int_array(GdamArg* arg)
{
	g_free(GDAM_ARG_INT_ARRAY(arg).array);
}
static void gdam_arg_clearer_int64_array(GdamArg* arg)
{
	g_free(GDAM_ARG_INT64_ARRAY(arg).array);
}
static void gdam_arg_clearer_string_array(GdamArg* arg)
{
	int i, n = GDAM_ARG_STRING_ARRAY(arg).length;
	for (i = 0; i < n; i++)
		g_free(GDAM_ARG_STRING_ARRAY(arg).array[i]);
	g_free(GDAM_ARG_STRING_ARRAY(arg).array);
}
static void gdam_arg_clearer_double_array(GdamArg* arg)
{
	g_free(GDAM_ARG_DOUBLE_ARRAY(arg).array);
}


/*
 * Initializing the fundamental-type hashtables.
 */
static GHashTable* arg_parser_types = NULL;
static GHashTable* type_to_info = NULL;

typedef struct _ArgTypeInfo ArgTypeInfo;
static struct _ArgTypeInfo {
	int type;
	ArgumentParser parser;
	ArgumentValist args;
	ArgumentWriter writer;
	ArgumentClearer clearer;
} arg_type_infos[] = {
	{ 
		GDAM_ARG_TYPE_INT,
		gdam_arg_parse_int,
		gdam_arg_valist_int,
		gdam_arg_writer_int,
		NULL
	},
	{
		GDAM_ARG_TYPE_INT64,
		gdam_arg_parse_int64,
		gdam_arg_valist_int64,
		gdam_arg_writer_int64,
		NULL
	},
	{
		GDAM_ARG_TYPE_STRING,
		gdam_arg_parse_string,
		gdam_arg_valist_string,
		gdam_arg_writer_string,
		gdam_arg_clearer_string
	},
	{
		GDAM_ARG_TYPE_DOUBLE,
		gdam_arg_parse_double,
		gdam_arg_valist_double,
		gdam_arg_writer_double,
		NULL
	},
	{
		GDAM_ARG_TYPE_INT_ARRAY,
		gdam_arg_parse_int_array,
		gdam_arg_valist_int_array,
		gdam_arg_writer_int_array,
		gdam_arg_clearer_int_array
	},
	{
		GDAM_ARG_TYPE_INT64_ARRAY,
		gdam_arg_parse_int64_array,
		gdam_arg_valist_int64_array,
		gdam_arg_writer_int64_array,
		gdam_arg_clearer_int64_array
	},
	{
		GDAM_ARG_TYPE_STRING_ARRAY,
		gdam_arg_parse_string_array,
		gdam_arg_valist_string_array,
		gdam_arg_writer_string_array,
		gdam_arg_clearer_string_array
	},
	{
		GDAM_ARG_TYPE_DOUBLE_ARRAY,
		gdam_arg_parse_double_array,
		gdam_arg_valist_double_array,
		gdam_arg_writer_double_array,
		gdam_arg_clearer_double_array
	}
};

#define NUM_ARG_TYPE_INFOS  (sizeof(arg_type_infos)/sizeof(arg_type_infos[0]))

void gdam_arg_init()
{
	int i;
	arg_parser_types = g_hash_table_new(NULL, NULL);
	for (i = 0; i < NUM_ARG_TYPE_INFOS; i++) {
		ArgTypeInfo* info = arg_type_infos + i;
		gpointer key = GINT_TO_POINTER(info->type);
		g_hash_table_insert(arg_parser_types, key, info);
	}
	type_to_info = g_hash_table_new(NULL, NULL);
}

/*=====================================================================*
 *                      gdam_arg_parse_from_buffer                     *
 *                                                                     *
 * Parse an array of arguments into a buffer.                          *
 *=====================================================================*/
GdamArg*      gdam_arg_parse_from_buffer(GdamBuffer* buffer,
                                         int num_args)
{
	GdamArg* rv;
	ArgTypeInfo* info;
	int i;
	g_return_val_if_fail(arg_parser_types != NULL, NULL);
	rv = g_new(GdamArg, num_args);
	for (i = 0; i < num_args; i++) {
		int type;
		rv[i].name = gdam_buffer_parse_string0(buffer);
		if (!rv[i].name) {
			gdam_debug("error parsing argument %d", i);
			break;
		}
		if (!gdam_buffer_parse_int(buffer, &type)) {
			gdam_debug("error parsing type of argument %s",
			           rv[i].name);
			g_free(rv[i].name);
			break;
		}
		rv[i].type = type;
		info = (ArgTypeInfo*)
				g_hash_table_lookup(arg_parser_types,
						    GINT_TO_POINTER(type));
		if (!info || !info->parser) {
			gdam_debug("no parser found for %s [type %d]",
			           rv[i].name, type);
			break;
		}
		if (!(*info->parser)(buffer, rv + i)) {
			g_free(rv[i].name);
			break;
		}
	}
	if (i == num_args)
		return rv;

	/* error occurred above... */
	gdam_arg_free_array(i, rv);
	return NULL;
}

GdamArg*      gdam_arg_parse_from_vararg(int num_args,
                                         va_list* args)
{
	GdamArg* rv;
	int i;
	g_return_val_if_fail(arg_parser_types != NULL, NULL);
	rv = g_new(GdamArg, num_args);
	for (i = 0; i < num_args; i++) {
		int type;
		ArgTypeInfo* info;
		rv[i].name = g_strdup(va_arg(*args, char*));
		if (!rv[i].name) {
			gdam_debug("error parsing argument %d", i);
			break;
		}
		type = va_arg(*args, int);
		rv[i].type = type;
		info = (ArgTypeInfo*) g_hash_table_lookup(arg_parser_types,
						    GINT_TO_POINTER(type));
		if (!info || !info->args) {
			gdam_debug("no parser found for %s [type %d]",
			           rv[i].name, type);
			break;
		}
		(*info->args)(args, rv + i);
	}
	if (i == num_args)
		return rv;

	/* error occurred above... */
	gdam_arg_free_array(i, rv);
	return NULL;
}
GdamArg*      gdam_arg_new_from_valist  (int             num_args,
                                         ...)
{
	va_list args;
	GdamArg* arg_rv;
	va_start(args, num_args);
	arg_rv = gdam_arg_parse_from_vararg(num_args, &args);
	va_end(args);
	return arg_rv;
}

void          gdam_arg_write_to_buffer  (GdamBuffer* buffer,
                                         int num_args,
					 const GdamArg* args)
{
	int i;
	g_return_if_fail(arg_parser_types != NULL);
	for (i = 0; i < num_args; i++) {
		int type = args[i].type;
		ArgTypeInfo* info;
		gdam_buffer_append_string0(buffer, args[i].name);
		gdam_buffer_append_int(buffer, type);
		info = (ArgTypeInfo*) g_hash_table_lookup(arg_parser_types,
						    GINT_TO_POINTER(type));
		if (!info || !info->writer) {
			gdam_debug("no writer found for %s [type %d]",
			           args[i].name, type);
			break;
		}
		(*info->writer)(buffer, args + i);
	}
}

void          gdam_arg_free_array       (int num_args,
                                         GdamArg* args)
{
	int i;
	for (i = 0; i < num_args; i++) {
		int type = args[i].type;
		ArgTypeInfo* info;
		g_free(args[i].name);
		info = (ArgTypeInfo*) g_hash_table_lookup(arg_parser_types,
						    GINT_TO_POINTER(type));
		if (info && info->clearer)
			(*info->clearer)(args + i);
	}
	g_free(args);
}


/*
 *  ____       _   _   _                               _
 * / ___|  ___| |_| |_(_)_ __   __ _    __ _ _ __   __| |
 * \___ \ / _ \ __| __| | '_ \ / _` |  / _` | '_ \ / _` |
 *  ___) |  __/ |_| |_| | | | | (_| | | (_| | | | | (_| |
 * |____/ \___|\__|\__|_|_| |_|\__, |  \__,_|_| |_|\__,_|
 *                             |___/
 *             _   _   _
 *   __ _  ___| |_| |_(_)_ __   __ _
 *  / _` |/ _ \ __| __| | '_ \ / _` |
 * | (_| |  __/ |_| |_| | | | | (_| |
 *  \__, |\___|\__|\__|_|_| |_|\__, |
 *  |___/                      |___/
 */
struct _GdamArgInfo {
	int		type;
	gboolean        is_function;
	GdamArgSetFunc  setter;
	GdamArgGetFunc  getter;
	GdamArgGenericFunc  function;
	unsigned	vital : 1,
	                trigger : 1;
};

typedef struct _GdamArgTrap GdamArgTrap;
typedef struct _GdamArgObjectInfo GdamArgObjectInfo;
struct _GdamArgTrap {
	guint		id;
	GdamArgInfo*	arg_info;
	GdamArgFunc	func;
	gpointer	data;
};

struct _GdamArgObjectInfo {
	GSList*		traps;
	guint		destroy_id;
	guint		last_trap_id;

	/* List of arginfos for currently pending traps.
	 */
	GSList*		pending_infos;
};

static guint gdam_arg_object_info_quark = 0;

static void make_quark()
{
	gdam_arg_object_info_quark = gtk_object_data_force_id("gdam-arg-data");
}

static void gdam_arg_object_info_free(GdamArgObjectInfo* info)
{
	g_slist_foreach(info->traps, (GFunc) g_free, NULL);
	g_slist_free(info->traps);
	g_free(info);
}

static GdamArgObjectInfo* gdam_arg_object_info_try(GtkObject* object)
{
	if (gdam_arg_object_info_quark == 0)
		return NULL;
	return gtk_object_get_data_by_id(object, gdam_arg_object_info_quark);
}
static GdamArgObjectInfo* gdam_arg_object_info_force(GtkObject* object)
{
	GdamArgObjectInfo* info;
	if (gdam_arg_object_info_quark == 0)
		make_quark();
	info = gtk_object_get_data_by_id(object, gdam_arg_object_info_quark);
	if (info == NULL) {
		info = g_new0(GdamArgObjectInfo, 1);
		info->destroy_id = gtk_signal_connect_object
						(object,
						 "destroy",
				 GTK_SIGNAL_FUNC(gdam_arg_object_info_free),
				                 (GtkObject*) info);
		gtk_object_set_data_by_id(object,
		                          gdam_arg_object_info_quark,
					  info);
	}
	return info;
}

void          gdam_arg_register         (GtkObjectClass* klass,
	                                 const char*     argname,
	                                 int             type,
					 GdamArgSetFunc  setter,
					 GdamArgGetFunc  getter,
					 int             gdam_arg_options)
{
	GHashTable* info_hash;
	GdamArgInfo* info;
	info_hash = (GHashTable*)g_hash_table_lookup(type_to_info,
					     GINT_TO_POINTER(klass->type));
	if (info_hash == NULL) {
		info_hash = g_hash_table_new(g_str_hash, g_str_equal);
		g_hash_table_insert(type_to_info,
		                    GINT_TO_POINTER(klass->type),
		                    info_hash);
	}

	info = (GdamArgInfo*) g_hash_table_lookup(info_hash, argname);
	g_return_if_fail(info == NULL);

	info = g_new(GdamArgInfo, 1);
	info->type = type;
	info->setter = setter;
	info->getter = getter;
	info->is_function = FALSE;
	info->function = NULL;
	if ((gdam_arg_options & GDAM_ARG_VITAL) == GDAM_ARG_VITAL)
		info->vital = 1;
	else
		info->vital = 0;
	if ((gdam_arg_options & GDAM_ARG_TRIGGER) == GDAM_ARG_TRIGGER)
		info->trigger = 1;
	else
		info->trigger = 0;

	g_hash_table_insert(info_hash, g_strdup(argname), info);
}

void          gdam_arg_function_register(GtkObjectClass* object_class,
	                                 const char*     function_name,
					 GdamArgGenericFunc  function)
{
	GHashTable* info_hash;
	GdamArgInfo* info;

	info_hash = (GHashTable*)g_hash_table_lookup(type_to_info,
				     GINT_TO_POINTER(object_class->type));
	if (info_hash == NULL) {
		info_hash = g_hash_table_new(g_str_hash, g_str_equal);
		g_hash_table_insert(type_to_info,
		                    GINT_TO_POINTER(object_class->type),
		                    info_hash);
	}

	info = (GdamArgInfo*) g_hash_table_lookup(info_hash, function_name);
	g_return_if_fail(info == NULL);

	info = g_new(GdamArgInfo, 1);
	info->type = 0;
	info->setter = NULL;
	info->getter = NULL;
	info->is_function = TRUE;
	info->function = function;
	info->vital = 0;
	info->trigger = 0;
	g_hash_table_insert(info_hash, g_strdup(function_name), info);
}

static GdamArgInfo* find_arg_info(GtkType         type,
                                  const char*     argname)
{
	GHashTable* info_hash;
	while (type) {
		info_hash = (GHashTable*)g_hash_table_lookup(
						type_to_info,
						GINT_TO_POINTER(type));
		if (info_hash != NULL) {
			GdamArgInfo* info = (GdamArgInfo*)
				g_hash_table_lookup(info_hash, argname);
			if (info)
				return info;
		}
		type = gtk_type_parent(type);
	}
	return NULL;
}


/*
 * Parse a human-readable value, given its type (inside the arg).
 */
static gboolean parse_int(const char* text, int* out)
{
	char* endp;
	*out = (int) strtol(text, &endp, 0);
	return endp != text && *endp == 0 ? TRUE : FALSE;
}

static gboolean parse_double(const char* text, gdouble* out)
{
	char* endp;
	*out = strtod(text, &endp);
	return endp != text && *endp == 0 ? TRUE : FALSE;
}

static gboolean parse_int64(const char* text, gint64* out)
{
	char* endp;
	*out = strtoll(text, &endp, 0);
	return endp != text && *endp == 0 ? TRUE : FALSE;
}

static gboolean parse_int_array(const char* text,
                                int*        length,
				int**       array)
{
	const char* at = text;
	int num_int = 0, max_int;
	while (*at) {
		while (*at && isspace(*at))
			at++;
		if (!*at)
			break;
		num_int++;
		while (*at && isxdigit(*at))
			at++;
	}
	if (*at)
		g_warning("unknown character %c in int array", *at);
	*length = num_int;
	*array = g_new(int, num_int);
	max_int = num_int;
	num_int = 0;
	while (*at && num_int < max_int) {
		while (*at && isspace(*at))
			at++;
		if (!*at)
			break;
		num_int++;
		*array[num_int++] = strtol(at, (char**) &at, 0);
	}
	return TRUE;
}

static gboolean parse_int64_array(const char* text,
                                int*        length,
				gint64**       array)
{
	const char* at = text;
	int num_int = 0, max_int;
	while (*at) {
		while (*at && isspace(*at))
			at++;
		if (!*at)
			break;
		num_int++;
		while (*at && isxdigit(*at))
			at++;
	}
	if (*at)
		g_warning("unknown character %c in int array", *at);
	*length = num_int;
	*array = g_new(gint64, num_int);
	max_int = num_int;
	num_int = 0;
	while (*at && num_int < max_int) {
		while (*at && isspace(*at))
			at++;
		if (!*at)
			break;
		num_int++;
		*array[num_int++] = strtoll(at, (char**) &at, 0);
	}
	return TRUE;
}

static gboolean parse_double_array(const char* text,
                                int*        length,
				gdouble**       array)
{
	const char* at = text;
	int num_int = 0, max_int;
	while (*at) {
		while (*at && isspace(*at))
			at++;
		if (!*at)
			break;
		num_int++;
		/* Yuck... how to do this portably... */
		while (*at && strchr("0123456789eE+-.", *at))
			at++;
	}
	if (*at)
		g_warning("unknown character %c in int array", *at);
	*length = num_int;
	*array = g_new(gdouble, num_int);
	max_int = num_int;
	num_int = 0;
	while (*at && num_int < max_int) {
		while (*at && isspace(*at))
			at++;
		if (!*at)
			break;
		num_int++;
		*array[num_int++] = strtod(at, (char**) &at);
	}
	return TRUE;
}

gboolean parse_string_array(const char* text, int* len, char*** array)
{
	char** at;
	int num;
	*array = g_strsplit(text, "\t", -1);
	if (*array == NULL) {
		*len = 0;
		return TRUE;
	}
	at = *array;
	num = 0;
	while (*at++)
		num++;
	*len = num;
	return TRUE;
}

gboolean gdam_arg_parse_val_from_string(GdamArg* arg, const char* text)
{
	switch (arg->type) {
		case GDAM_ARG_TYPE_INT:
			return parse_int(text, &GDAM_ARG_INT(arg));
		case GDAM_ARG_TYPE_DOUBLE:
			return parse_double(text, &GDAM_ARG_DOUBLE(arg));
		case GDAM_ARG_TYPE_STRING:
			GDAM_ARG_STRING(arg) = g_strdup(text);
			return TRUE;
		case GDAM_ARG_TYPE_INT64:
			return parse_int64(text, &GDAM_ARG_INT64(arg));
		case GDAM_ARG_TYPE_INT_ARRAY:
			return parse_int_array(text,
					&GDAM_ARG_INT_ARRAY(arg).length,
					&GDAM_ARG_INT_ARRAY(arg).array);
		case GDAM_ARG_TYPE_DOUBLE_ARRAY:
			return parse_double_array(text,
					&GDAM_ARG_DOUBLE_ARRAY(arg).length,
					&GDAM_ARG_DOUBLE_ARRAY(arg).array);
		case GDAM_ARG_TYPE_STRING_ARRAY:
			return parse_string_array(text,
					&GDAM_ARG_STRING_ARRAY(arg).length,
					&GDAM_ARG_STRING_ARRAY(arg).array);
		case GDAM_ARG_TYPE_INT64_ARRAY:
			return parse_int64_array(text,
					&GDAM_ARG_INT64_ARRAY(arg).length,
					&GDAM_ARG_INT64_ARRAY(arg).array);
		default:
			g_warning("unknown argument type [%d]", arg->type);
			return FALSE;
	}
	g_assert_not_reached();
	return FALSE;
}

char* gdam_arg_val_as_string(GdamArg* arg)
{
	switch(arg->type) {
		case GDAM_ARG_TYPE_INT:
			return g_strdup_printf("%d", GDAM_ARG_INT(arg));
		case GDAM_ARG_TYPE_INT64:
			return g_strdup_printf("%lldL", GDAM_ARG_INT64(arg));
		case GDAM_ARG_TYPE_DOUBLE:
			return g_strdup_printf("%.6g", GDAM_ARG_DOUBLE(arg));
		case GDAM_ARG_TYPE_STRING:
			return g_strdup(GDAM_ARG_STRING(arg));
		case GDAM_ARG_TYPE_INT_ARRAY: {
			char buf[40];
			char* val;
			int i;
			int length = GDAM_ARG_INT_ARRAY(arg).length;
			int* array = GDAM_ARG_INT_ARRAY(arg).array;
			GString* string = g_string_new("[");
			for (i = 0; i < length; i++) {
				snprintf(buf, sizeof(buf), "%d", array[i]);
				buf[sizeof(buf)-1] = '\0';
				if (i > 0)
					string = g_string_append_c(string,' ');
				string = g_string_append(string, buf);
			}
			string = g_string_append_c(string,']');
			val = string->str;
			g_string_free(string, FALSE);
			return val;
		}
		case GDAM_ARG_TYPE_INT64_ARRAY: {
			char buf[40];
			char* val;
			int i;
			int length = GDAM_ARG_INT64_ARRAY(arg).length;
			gint64* array = GDAM_ARG_INT64_ARRAY(arg).array;
			GString* string = g_string_new("[");
			for (i = 0; i < length; i++) {
				snprintf(buf, sizeof(buf), "%lld", array[i]);
				buf[sizeof(buf)-1] = '\0';
				if (i > 0)
					string = g_string_append_c(string,' ');
				string = g_string_append(string, buf);
			}
			string = g_string_append_c(string,']');
			val = string->str;
			g_string_free(string, FALSE);
			return val;
		}
		case GDAM_ARG_TYPE_DOUBLE_ARRAY: {
			char buf[40];
			int i;
			int length = GDAM_ARG_DOUBLE_ARRAY(arg).length;
			char* val;
			double* array = GDAM_ARG_DOUBLE_ARRAY(arg).array;
			GString* string = g_string_new("[");
			for (i = 0; i < length; i++) {
				snprintf(buf, sizeof(buf), "%.6g", array[i]);
				buf[sizeof(buf)-1] = '\0';
				if (i > 0)
					string = g_string_append_c(string,' ');
				string = g_string_append(string, buf);
			}
			string = g_string_append_c(string,']');
			val = string->str;
			g_string_free(string, FALSE);
			return val;
		}
		case GDAM_ARG_TYPE_STRING_ARRAY: {
			int i;
			int length = GDAM_ARG_STRING_ARRAY(arg).length;
			char** array = GDAM_ARG_STRING_ARRAY(arg).array;
			GString* string = g_string_new("[");
			char* val;
			for (i = 0; i < length; i++) {
				if (i > 0)
					string = g_string_append_c(string,' ');
				g_string_append_c(string, '"');
				g_string_append(string, array[i]);
				string = g_string_append_c(string, '"');
			}
			string = g_string_append_c(string,']');
			val = string->str;
			g_string_free(string, FALSE);
			return val;
		}
	}
	g_warning("type %d unknown", arg->type);
	return NULL;
}

char* gdam_arg_array_as_string(int num_args, GdamArg* arg)
{
	char** args;
	char* rv;
	int i;
	args = (char**) alloca(sizeof(char*) * num_args * 2 + 1);
	args[num_args * 2] = NULL;
	for (i = 0; i < num_args; i++) {
		args[2*i] = arg[i].name;
		args[2*i+1] = gdam_arg_val_as_string(arg + i);
	}
	rv = g_strjoinv(" ", args);
	for (i = 0; i < num_args; i++)
		g_free(args[2*i+1]);
	return rv;
}
	
gchar*   g_strjoinv		(const gchar  *separator,
				 gchar       **str_array);



static void gdam_arg_set_internal(GdamArgInfo*      info,
                                  GtkObject*        object,
				  GdamArg*          arg)
{
	GdamArgObjectInfo* object_info;
	object_info = gdam_arg_object_info_force(object);

	if (object_info != NULL) {
		if (g_slist_find(object_info->pending_infos, info))
			return;
	}

	object_info->pending_infos
		= g_slist_prepend(object_info->pending_infos, info);
	(*info->setter)(object, arg);
	if (object_info != NULL) {
		GSList* at;
		for (at = object_info->traps;
		     at != NULL;
		     at = at->next) {
		     	GdamArgTrap* trap = at->data;
			if (trap->arg_info == info) {
				(*trap->func)(trap->data, arg, object);
			}
		}
	}
	object_info->pending_infos
		= g_slist_remove(object_info->pending_infos, info);
}

void          gdam_arg_set_object_arg   (GtkObject*      object,
                                         GdamArg*        arg)
{
	GdamArgInfo* info;
	info = find_arg_info(object->klass->type, arg->name);
	if (!info || info->is_function || !info->setter) {
		gdam_debug("no set method for %s in %s", 
			arg->name, gtk_type_name(object->klass->type));
		return;
	}
	if (arg->type != info->type) {
		gdam_debug("set_object_arg: type mismatch in %s", arg->name);
		return;
	}
	if (gdam_debugging_on) {
		char* val;
		val = gdam_arg_val_as_string(arg);
		gdam_debug("set `%s' in `%s' [%p] to %s",
				      arg->name,
				      gtk_type_name(object->klass->type),
				      object,
				      val);
		g_free(val);
	}
	gdam_arg_set_internal(info, object, arg);
}

void          gdam_arg_set              (GtkObject*      object,
                                         const char*     name,
					 /* value... */
					 ...)
{
	GdamArgInfo* info;
	GdamArg arg;
	ArgTypeInfo* type_info;
	arg.name = (char*) name;
	info = find_arg_info(object->klass->type, arg.name);
	if (!info || info->is_function || !info->setter) {
		gdam_debug("no set method for %s in %s", 
			arg.name, gtk_type_name(object->klass->type));
		return;
	}
	arg.type = info->type;
	{
		va_list arg_list;
		type_info = g_hash_table_lookup(arg_parser_types,
						GINT_TO_POINTER(info->type));
		if (!type_info || !type_info->args) {
			gdam_debug("no parser found for %s [type %d]",
			           name, info->type);
			return;
		}
		va_start(arg_list, name);
		(*type_info->args)(arg_list, &arg);
		va_end(arg_list);
	}

	if (gdam_debugging_on) {
		char* val;
		val = gdam_arg_val_as_string(&arg);
		gdam_debug("set `%s' in `%s' [%p] to %s",
				      arg.name,
				      gtk_type_name(object->klass->type),
				      object,
				      val);
		g_free(val);
	}

	gdam_arg_set_internal(info, object, &arg);

	if (type_info->clearer != NULL)
		(*type_info->clearer)(&arg);

	return;
}

gboolean      gdam_arg_get_object_arg   (GtkObject*      object,
                                         GdamArg*        arg)
{
	GdamArgInfo* info;
	info = find_arg_info(object->klass->type, arg->name);
	if (!info || info->is_function || !info->getter) {
		gdam_debug("no get method for %s in %s", 
			arg->name, gtk_type_name(object->klass->type));
		return FALSE;
	}
	arg->type = info->type;
	(*info->getter)(object, arg);
	return TRUE;
}

int gdam_arg_query_type(GtkType type, const char* argname)
{
	GdamArgInfo* info;
	info = find_arg_info(type, argname);
	if (info == NULL)
		return 0;
	return info->type;
}


void          gdam_arg_get_by_name      (GtkObject*      object,
                                         int             num_args,
					 const char*const* arg_names,
					 int*            num_args_out,
					 GdamArg**       args_out)
{
	int i;
	GdamArg* args;
	*num_args_out = 0;

	*args_out = args = g_new(GdamArg, num_args);
	for (i = 0; i < num_args; i++) {
		GdamArg* arg = & args[*num_args_out];
		arg->name = g_strdup(arg_names[i]);
		if (gdam_arg_get_object_arg(object, arg))
			(*num_args_out)++;
	}
}

gboolean      gdam_arg_call_function    (GtkObject*      object,
                                         const char*     function_name,
					 int             num_args,
					 GdamArg*        args,
					 int*            num_rv_out,
					 GdamArg**       rv_out)
{
	GdamArgInfo* info;
	info = find_arg_info(object->klass->type, function_name);
	if (!info || !info->is_function || info->function == NULL) {
		gdam_debug("no function for %s in %s", 
			function_name, gtk_type_name(object->klass->type));
		return FALSE;
	}
	*num_rv_out = 0;
	*rv_out = NULL;
	(*info->function)(object, num_args, args, num_rv_out, rv_out);
	return TRUE;
}


typedef struct _GetAllArgsInfo GetAllArgsInfo;
struct _GetAllArgsInfo {
	GHashTable*		args_out_table;
	gboolean		only_vital;
	GtkObject*		object;
};
static void append_args(gpointer key, gpointer value, gpointer user_data)
{
	GetAllArgsInfo* get_info = (GetAllArgsInfo*)user_data;
	const char* arg_name = (const char*)key;
	GdamArgInfo* arg_info = (GdamArgInfo*)value;
	GdamArg* new_arg;

	/* Don't append "trigger" arguments. */
	if (arg_info->trigger)
		return;

	/* Non-vital argument... */
	if (get_info->only_vital && !arg_info->vital)
		return;

	/* Derived-classes get to override the information in their
	 * base-classes. */
	if (g_hash_table_lookup(get_info->args_out_table, arg_name))
		return;

	/* If there is no get function, we are unable to do this as well. */
	if (arg_info->getter == NULL)
		return;
	
	new_arg = g_new(GdamArg, 1);
	new_arg->name = g_strdup(arg_name);
	new_arg->type = arg_info->type;
	(*arg_info->getter)(get_info->object, new_arg);
	g_hash_table_insert(get_info->args_out_table, new_arg->name, new_arg);
}


static void copy_and_free(gpointer key, gpointer value, gpointer user_data)
{
	GdamArg** arg_at = (GdamArg**) user_data;
	**arg_at = *(GdamArg*)value;
	*arg_at = (*arg_at) + 1;
	g_free(value);
}
void          gdam_arg_get_all     (GtkObject*      object,
				    gboolean        only_vital,
                                    int*            num_args_out,
			            GdamArg**       args_out)
{
	GtkType type = object->klass->type;
	GHashTable* info_hash;
	GHashTable* args_out_table;
	GetAllArgsInfo info;
	GdamArg* arg_at;
	args_out_table = g_hash_table_new(g_str_hash, g_str_equal);
	info.args_out_table = args_out_table;
	info.only_vital = only_vital;
	info.object = object;
	while (type) {
		info_hash = (GHashTable*)g_hash_table_lookup(type_to_info,
					           GINT_TO_POINTER(type));
		if (info_hash != NULL)
			g_hash_table_foreach(info_hash,
			                     append_args,
					     &info);
		type = gtk_type_parent(type);
	}

	*args_out = g_new(GdamArg, g_hash_table_size(args_out_table));
	arg_at = *args_out;
	g_hash_table_foreach(args_out_table, copy_and_free, &arg_at);
	g_hash_table_destroy(args_out_table);

	*num_args_out = arg_at - *args_out;
	*args_out = g_renew(GdamArg, *args_out, *num_args_out);
}

guint         gdam_arg_connect          (GtkObject*      object,
                                         const char*     name,
					 GdamArgFunc     func,
					 gpointer        user_data)
{
	GdamArgInfo* arg_info;
	GdamArgObjectInfo* object_info;
	GdamArgTrap* trap;
	arg_info = find_arg_info(object->klass->type, name);
	if (arg_info == NULL) {
		g_warning("argument %s was not found for class %s",
				name, gtk_type_name(object->klass->type));
		return 0;
	}
	object_info = gdam_arg_object_info_force(object);
	trap = g_new(GdamArgTrap, 1);
	trap->id = ++(object_info->last_trap_id);
	trap->arg_info = arg_info;
	trap->func = func;
	trap->data = user_data;
	object_info->traps = g_slist_prepend(object_info->traps, trap);
	return trap->id;
}

void          gdam_arg_disconnect       (GtkObject*      object,
                                         guint           connect_id)
{
	GSList* at;
	GdamArgObjectInfo* object_info;
	g_return_if_fail(connect_id != 0);
	object_info = gdam_arg_object_info_try(object);
	g_return_if_fail(object_info != NULL);
	for (at = object_info->traps; at != NULL; at = at->next) {
		GdamArgTrap* trap = at->data;
		if (trap->id == connect_id)
			break;
	}
	g_return_if_fail(at != NULL);
	g_free(at->data);
	object_info->traps = g_slist_remove(object_info->traps, at->data);
}
